@@ -1,24 +1,35 @@
-from zope.interface import implements
+from zope.interface import implements, Interface
+
+from twisted.protocols.basic import LineReceiver
+LineReceiver.MAX_LENGTH = 1024*1024*64
 
 from twisted.internet import defer, reactor, protocol
-from twisted.web.client import Agent
 from twisted.web.http_headers import Headers
 from twisted.web.iweb import IBodyProducer
+from twisted.python import log
 
 from distutils.version import StrictVersion
 
 import urllib
 import re, csv
+import time
 from cStringIO import StringIO
 
 from xml.etree import ElementTree
 
 # MD_ resources
 from riakasaurus.metadata import *
+from twisted.web.client import Agent
+
 
 from riakasaurus.riak_index_entry import RiakIndexEntry
 from riakasaurus.mapreduce import RiakLink
 
+# protobuf
+from riakasaurus.tx_riak_pb import RiakPBCClient
+from riakasaurus.riak_kv_pb2 import *
+from riakasaurus.riak_pb2 import *
+
 MAX_LINK_HEADER_SIZE = 8192 - 8
 
 versions = {
@@ -27,6 +38,60 @@ versions = {
     1.2: StrictVersion("1.2.0")
     }
 
+
+class ITransport(Interface):
+    def get_keys(self, bucket):
+        """
+        list keys for a given bucket
+        """
+        
+    def put(self, robj, w = None, dw = None, pw = None, return_body = True, if_none_match=False):
+        """
+        store a riak_object
+        """
+
+    def get(self, robj, r = None, pr = None, vtag = None):
+        """
+        fetch a key from the server
+        """
+
+    def delete(self, robj, rw=None, r = None, w = None, dw = None, pr = None, pw = None):
+        """
+        delete a key from the bucket
+        """
+        
+    def server_version(self):
+        """
+        return cached server version 
+        """
+
+    def _server_version(self):
+        """
+        Gets the server version from the server. To be implemented by
+        the individual transport class.
+        :rtype string
+        """
+    def get_buckets(self):
+        """
+        return the existing buckets
+        """
+
+    def ping(self):
+        """
+        Check server is alive
+        """
+
+    def set_bucket_props(self, bucket, props):
+        """
+        Set bucket properties
+        """
+
+    def get_bucket_props(self, bucket):
+        """
+        get bucket properties
+        """
+        
+        
 class FeatureDetection(object):
     _s_version = None
 
@@ -44,7 +109,7 @@ class FeatureDetection(object):
         Whether MapReduce requests can be submitted without phases.
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1.1])
 
     @defer.inlineCallbacks
@@ -55,7 +120,7 @@ class FeatureDetection(object):
 
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1.2])
 
     @defer.inlineCallbacks
@@ -64,7 +129,7 @@ class FeatureDetection(object):
         Whether search queries are supported over Protocol Buffers
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1.2])
 
     @defer.inlineCallbacks
@@ -74,7 +139,7 @@ class FeatureDetection(object):
         Protocol Buffers
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1])
 
     @defer.inlineCallbacks
@@ -84,7 +149,7 @@ class FeatureDetection(object):
         e.g. primary quorums, basic_quorum, notfound_ok
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1])
 
     @defer.inlineCallbacks
@@ -93,7 +158,7 @@ class FeatureDetection(object):
         Whether 'not found' responses might include vclocks
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1])
 
     @defer.inlineCallbacks
@@ -103,7 +168,7 @@ class FeatureDetection(object):
         supported over Protocol Buffers
         :rtype bool
         """
-        d = yield self.server_version() 
+        d = yield self.server_version()
         defer.returnValue(d >= versions[1])
 
     @defer.inlineCallbacks
@@ -116,7 +181,7 @@ class FeatureDetection(object):
 class BodyReceiver(protocol.Protocol):
     """ Simple buffering consumer for body objects """
     def __init__(self, finished):
-        self.finished = finished 
+        self.finished = finished
         self.buffer = StringIO()
 
     def dataReceived(self, buffer):
@@ -147,9 +212,15 @@ class StringProducer(object):
         pass
 
 class HTTPTransport(FeatureDetection):
+
+    implements(ITransport)
+    
     """ HTTP Transport for Riak """
-    def __init__(self, client):
-        self._prefix = client._prefix
+    def __init__(self, client, prefix=None):
+        if prefix:
+            self._prefix = prefix
+        else:
+            self._prefix = client._prefix
         self.host = client._host
         self.port = client._port
         self.client = client
@@ -160,7 +231,7 @@ class HTTPTransport(FeatureDetection):
             headers = {"http_code": response.code}
             for key, val in response.headers.getAllRawHeaders():
                 headers[key.lower()] = val[0]
-            
+
             return headers, body.read()
 
         if response.length:
@@ -176,9 +247,12 @@ class HTTPTransport(FeatureDetection):
         h = {}
         for k, v in headers.items():
             if not isinstance(v, list):
-                h[k] = [v]
+                h[k.lower()] = [v]
             else:
-                h[k] = v 
+                h[k.lower()] = v
+
+        if not 'content-type' in h.keys():
+            h['content-type'] = ['application/json']
 
         if body:
             bodyProducer = StringProducer(body)
@@ -228,14 +302,14 @@ class HTTPTransport(FeatureDetection):
     def get_keys(self, bucket):
         params = {'props' : 'True', 'keys' : 'true'}
         url = self.build_rest_path(bucket, params=params)
-        
+
 
         headers, encoded_props = yield self.http_request('GET', url)
 
         if headers['http_code'] == 200:
             props = self.decodeJson(encoded_props)
         else:
-            raise Exception('Error getting bucket properties.') 
+            raise Exception('Error getting bucket properties.')
 
         defer.returnValue(props['keys'])
 
@@ -243,7 +317,7 @@ class HTTPTransport(FeatureDetection):
     def set_bucket_props(self, bucket, props):
         """
         Set bucket properties
-        """ 
+        """
         url = self.build_rest_path(bucket)
         headers = {'Content-Type': 'application/json'}
         content = self.encodeJson({'props': props})
@@ -299,11 +373,11 @@ class HTTPTransport(FeatureDetection):
         # If stats is disabled, we can't assume the Riak version
         # is >= 1.1. However, we can assume the new URL scheme is
         # at least version 1.0
-        elif 'riak_kv_wm_buckets' in self.get_resources():
+        elif 'riak_kv_wm_buckets' in (yield self.get_resources()):
             defer.returnValue("1.0.0")
         else:
             defer.returnValue("0.14.0")
-   
+
     @defer.inlineCallbacks
     def get_resources(self):
         """
@@ -311,7 +385,7 @@ class HTTPTransport(FeatureDetection):
         :rtype dict
         """
         response = yield self.http_request('GET', '/', {'Accept':'application/json'})
-        if response[0]['http_status'] is 200:
+        if response[0]['http_code'] is 200:
             defer.returnValue(self.decodeJson(response[1]))
         else:
             defer.returnValue({})
@@ -333,6 +407,24 @@ class HTTPTransport(FeatureDetection):
             self.parse_body(response, [200, 300, 404])
         )
 
+    @defer.inlineCallbacks
+    def head(self, robj, r = None, pr = None, vtag = None) :
+        """
+        Get metadata for a bucket/key from the server, basically
+        the same as get() but retrieves no data
+        """
+        params = {'r' : r, 'pr': pr}
+        if vtag is not None:
+            params['vtag'] = vtag
+        url = self.build_rest_path(robj.get_bucket(), robj.get_key(),
+                                   params=params)
+
+        response = yield self.http_request('HEAD', url)
+        defer.returnValue(
+            self.parse_body(response, [200, 300, 404])
+        )
+        
+        
     def put(self, robj, w = None, dw = None, pw = None, return_body = True, if_none_match=False):
         """
         Serialize put request and deserialize response
@@ -343,6 +435,7 @@ class HTTPTransport(FeatureDetection):
         url = self.build_rest_path(bucket=robj.get_bucket(), key=robj.get_key(),
                                    params=params)
         headers = self.build_put_headers(robj)
+
         # TODO: use a more general 'prevent_stale_writes' semantics,
         # which is a superset of the if_none_match semantics.
         if if_none_match:
@@ -470,6 +563,7 @@ class HTTPTransport(FeatureDetection):
         result = self.decodeJson(response[1])
         defer.returnValue(result)
 
+    @defer.inlineCallbacks
     def get_index(self, bucket, index, startkey, endkey=None):
         """
         Performs a secondary index query.
@@ -479,11 +573,13 @@ class HTTPTransport(FeatureDetection):
         if endkey:
             segments.append(str(endkey))
         uri = '/%s' % ('/'.join(segments))
-        headers, data = response = self.get_request(uri)
+        headers, data = response = yield self.get_request(uri)
         self.check_http_code(response, [200])
         jsonData = self.decodeJson(data)
-        return jsonData[u'keys'][:]
 
+        defer.returnValue(jsonData[u'keys'][:])
+
+    @defer.inlineCallbacks
     def search(self, index, query, **params):
         """
         Performs a search query.
@@ -499,13 +595,13 @@ class HTTPTransport(FeatureDetection):
         options.update(params)
         # TODO: use resource detection
         uri = "/solr/%s/select" % index
-        headers, data = response = self.get_request(uri, options)
+        headers, data = response = yield self.get_request(uri, options)
         self.check_http_code(response, [200])
         if 'json' in headers['content-type']:
             results = self.decodeJson(data)
-            return self._normalize_json_search_response(results)
+            defer.returnValue(self._normalize_json_search_response(results))
         elif 'xml' in headers['content-type']:
-            return self._normalize_xml_search_response(data)
+            defer.returnValue(self._normalize_xml_search_response(data))
         else:
             raise ValueError("Could not decode search response")
 
@@ -561,7 +657,7 @@ class HTTPTransport(FeatureDetection):
             elif header == 'charset':
                 metadata[MD_CHARSET] = value
             elif header == 'content-encoding':
-                metadata[MD_CTYPE] = value
+                metadata[MD_ENCODING] = value
             elif header == 'etag':
                 metadata[MD_VTAG] = value
             elif header =='link':
@@ -621,7 +717,6 @@ class HTTPTransport(FeatureDetection):
             for link in links:
                 header = self.to_link_header(link)
                 if len(current_header + header) > MAX_LINK_HEADER_SIZE:
-                    headers.add('Link', current_header)
                     current_header = ''
 
                 if current_header != '': header = ', ' + header
@@ -825,5 +920,342 @@ class XMLSearchResult(object):
                 'max_score':self.max_score,
                 'docs': self.docs }
 
+class StatefulTransport(object):
+
+    def __init__(self,transport):
+        self.__transport = transport
+        self.__state = 'idle'
+        self.__created = time.time()
+        self.__used = time.time()
+
+    def __repr__(self):
+        return '<StatefulTransport idle=%.2fs state=\'%s\' transport=%s>' % (time.time() - self.__used, self.__state, self.__transport)
+
+    def isActive(self):
+        return self.__state == 'active'
+
+    def setActive(self):
+        self.__state = 'active'
+        self.__used = time.time()
+
+    def isIdle(self):
+        return self.__state == 'idle'
+
+    def setIdle(self):
+        self.__state = 'idle'
+        self.__used = time.time()
+
+    def getTransport(self):
+        return self.__transport
+
+    def age(self):
+        return time.time() - self.__used
+        
+
+class PBCTransport(FeatureDetection):
+    """ Protocoll buffer transport for Riak """
+
+    implements(ITransport)
+
+    debug = 0
+    MAX_TRANSPORTS = 50
+    MAX_IDLETIME   = 5*60     # in seconds
+    GC_TIME        = 100        # how often (in seconds) the garbage collection should run
+    timeout        = None
+    
+    def __init__(self, client):
+        self._prefix = client._prefix
+        self.host = client._host
+        self.port = client._port
+        self.client = client
+        self._client_id = None
+        self._transports = []    # list of transports, empty on start
+        self._gc = reactor.callLater(self.GC_TIME, self._garbageCollect)
+
+    def setTimeout(self,t):
+        self.timeout = t
+
+    @defer.inlineCallbacks
+    def _getFreeTransport(self):
+        foundOne = False
+        for stp in self._transports:
+            if stp.isIdle():
+                stp.setActive()
+                foundOne = True
+                if self.debug:
+                    print "[%s] aquired idle transport[%d]: %s" % (self.__class__.__name__, len(self._transports),stp)
+                defer.returnValue(stp)
+        if not foundOne:
+            if len(self._transports) > self.MAX_TRANSPORTS:
+                raise Exception("to many transports, aborting")
+
+            # nothin free, create a new protocol instance, append
+            # it to self._transports and return it
+            transport = yield RiakPBCClient().connect(self.host, self.port)
+            if self.timeout:
+                transport.setTimeout(self.timeout)
+            stp = StatefulTransport(transport)
+            idx = len(self._transports)
+            self._transports.append(stp)
+            stp.setActive()
+            if self.debug:
+                print "[%s] allocate new transport[%d]: %s" % (self.__class__.__name__, len(self._transports),stp)
+            defer.returnValue(stp)
+
+    @defer.inlineCallbacks
+    def _garbageCollect(self):
+        for idx, stp in enumerate(self._transports):
+            if stp.isIdle() and stp.age() > self.MAX_IDLETIME:
+                yield stp.getTransport().quit()
+                if self.debug:
+                    print "[%s] expire transport[%d] %s" % (self.__class__.__name__, idx,stp)
+                del self._transports[idx]
+        self._gc = reactor.callLater(self.GC_TIME, self._garbageCollect)
+        
+    @defer.inlineCallbacks
+    def quit(self):
+        self._gc.cancel()      # cancel the garbage collector
+
+        for stp in self._transports:
+            if self.debug:
+                print "[%s] transport[%d].quit() %s" % (self.__class__.__name__, len(self._transports),stp)
+            yield stp.getTransport().quit()
+        
+    def __del__(self):
+        """on shutdown, close all transports"""
+        self.quit()
+
+    @defer.inlineCallbacks
+    def put(self, robj, w = None, dw = None, pw = None, return_body = True, if_none_match=False):
+        # std kwargs
+        kwargs = {'w'             : w,
+                  'dw'            : dw,
+                  'pw'            : pw,
+                  'return_body'   : return_body,
+                  'if_none_match' : if_none_match
+                  }
+        # vclock
+        vclock = robj.vclock() or None
+
+        payload = {
+            'value' : robj.get_encoded_data(),
+            'content_type' : robj.get_content_type(),
+            }
+        
+        # links
+        links = robj.get_links()
+        if links:
+            payload['links'] = []
+            for l in links:
+                payload['links'].append((l.get_bucket(), l.get_key(), l.get_tag()))
+
+        # usermeta
+        if robj.get_usermeta():
+            payload['usermeta'] = []
+            for key, value in robj.get_usermeta().iteritems():
+                payload['usermeta'].append((key, value))
+
+                
+        # aquire transport, fire, release
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.put(robj.get_bucket().get_name(),
+                                  robj.get_key(),
+                                  payload,
+                                  vclock,
+                                  **kwargs
+                                  )
+        stp.setIdle()
+        if return_body:
+            # defer.returnValue([x.value for x in ret.content])
+            defer.returnValue(self.parseRpbGetResp(ret))
+        else:
+            defer.returnValue(None)
+
+    @defer.inlineCallbacks
+    def get(self, robj, r = None, pr = None, vtag = None):
+
+        # ***FIXME*** whats vtag for? ignored for now
+        
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.get(robj.get_bucket().get_name(),
+                                  robj.get_key(),
+                                  r = r,
+                                  pr = pr)
+
+        stp.setIdle()
+        defer.returnValue(self.parseRpbGetResp(ret))
+
+    @defer.inlineCallbacks
+    def head(self, robj, r = None, pr = None, vtag = None):
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.get(robj.get_bucket().get_name(),
+                                  robj.get_key(),
+                                  r = r,
+                                  pr = pr,
+                                  head = True)
+
+        stp.setIdle()
+        defer.returnValue(self.parseRpbGetResp(ret))
 
 
+    @defer.inlineCallbacks
+    def delete(self, robj, rw=None, r = None, w = None, dw = None, pr = None, pw = None):
+        """
+        Delete an object.
+        """
+        # We could detect quorum_controls here but HTTP ignores
+        # unknown flags/params.
+        kwargs = {'rw' : rw, 'r': r, 'w': w, 'dw': dw, 'pr': pr, 'pw': pw}
+        headers = {}
+
+        ts = yield self.tombstone_vclocks()
+        if ts and robj.vclock() is not None:
+            kwargs['vclock'] = robj.vclock()
+
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.delete(robj.get_bucket().get_name(),
+                                     robj.get_key(),
+                                     **kwargs
+                                     )
+
+        stp.setIdle()
+        defer.returnValue(ret)
+
+
+    @defer.inlineCallbacks
+    def get_buckets(self):
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.getBuckets()
+        stp.setIdle()
+        defer.returnValue([x for x in ret.buckets])
+
+        
+    @defer.inlineCallbacks
+    def server_version(self):
+        if not self._s_version:
+            self._s_version = yield self._server_version()
+
+        defer.returnValue(StrictVersion(self._s_version))
+
+    @defer.inlineCallbacks
+    def _server_version(self):
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+
+        stats = yield transport.getServerInfo()
+        stp.setIdle()
+
+        
+        if stats is not None:
+            if self.debug:
+                print "[%s] fetched server version: %s" % (self.__class__.__name__, stats.server_version)
+            defer.returnValue(stats.server_version)
+        else:
+            defer.returnValue("0.14.0")
+
+    @defer.inlineCallbacks
+    def ping(self):
+        """
+        Check server is alive
+        """
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.ping()
+        stp.setIdle()
+        defer.returnValue(ret == True)
+
+
+    @defer.inlineCallbacks
+    def set_bucket_props(self, bucket, props):
+        """
+        Set bucket properties
+        """
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.setBucketProperties(bucket.get_name(), **props)
+        stp.setIdle()
+        defer.returnValue(ret == True)
+
+
+    @defer.inlineCallbacks
+    def get_bucket_props(self, bucket):
+        """
+        get bucket properties
+        """
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.getBucketProperties(bucket.get_name())
+        stp.setIdle()
+        defer.returnValue({'n_val'      : ret.props.n_val,
+                           'allow_mult' : ret.props.allow_mult})
+        
+    @defer.inlineCallbacks
+    def get_keys(self, bucket):
+        stp = yield self._getFreeTransport()
+        transport = stp.getTransport()
+        ret = yield transport.getKeys(bucket.get_name())
+        stp.setIdle()
+        defer.returnValue(ret)
+
+    def parseRpbGetResp(self,res):
+        """
+        adaptor for a RpbGetResp message
+        message RpbGetResp {
+           repeated RpbContent content = 1;
+           optional bytes vclock = 2;
+           optional bool unchanged = 3;
+        }
+        """
+        if res == True:         # empty response
+            return None
+        vclock = res.vclock
+        resList = []
+        for content in res.content: # iterate over RpbContent field
+            metadata = {MD_USERMETA: {}, MD_INDEX: []}
+            data = content.value
+            if content.HasField('content_type'): metadata[MD_CTYPE] = content.content_type
+            if content.HasField('charset'): metadata[MD_CHARSET] = content.charset
+            if content.HasField('content_encoding'): metadata[MD_ENCODING] = content.content_encoding
+            if content.HasField('vtag'): metadata[MD_VTAG] = content.vtag
+            if content.HasField('last_mod'): metadata[MD_LASTMOD] = content.last_mod
+            if content.HasField('deleted'): metadata[MD_DELETED] = content.deleted
+
+            if len(content.links):
+                metadata[MD_LINKS] = []
+                for l in content.links:
+                    metadata[MD_LINKS].append(RiakLink(l.bucket,l.key,l.tag))
+
+            if len(content.usermeta):
+                metadata[MD_USERMETA] = {}
+                for md in content.usermeta:
+                    metadata[MD_USERMETA][md.key] = md.value
+                    
+            if len(content.indexes):
+                metadata[MD_INDEX] = []
+                for ie in content.indexes:
+                    metadata[MD_INDEX].append(RiakIndexEntry(ie.key, ie.value))
+            resList.append((metadata, data))
+        return vclock, resList
+
+            
+    def decodeJson(self, s):
+        return self.client.get_decoder('application/json')(s)
+
+    def encodeJson(self, s):
+        return self.client.get_encoder('application/json')(s)
+        
+    # def deferred_sleep(self,secs):
+    #     """
+    #     fake deferred sleep
+        
+    #     @param secs: time to sleep
+    #     @type secs: float
+    #     """
+    #     d = defer.Deferred()
+    #     reactor.callLater(secs, d.callback, None)
+    #     return d
